@@ -8,16 +8,34 @@ from __future__ import unicode_literals
 from gi.repository import Gio
 from gi.repository import Gtk
 
-from .async_ import Async
+from .async_ import Async, Coroutine, Return
 from .common import setdefault
 from .compat import basestring
 from .locale import _
-from .mount import Action, Branch
+from .mount import Action, prune_empty_node
+from .prompt import Dialog
 
 
 __all__ = ['UdiskieMenu', 'SmartUdiskieMenu', 'TrayIcon']
 
 
+class MenuFolder(object):
+
+    def __init__(self, label, items):
+        self.label = label
+        self.items = items
+
+    def __bool__(self):
+        return bool(self.items)
+
+    __nonzero__ = __bool__
+
+
+class MenuSection(MenuFolder): pass
+class SubMenu(MenuFolder): pass
+
+
+
 class Icons(object):
 
     """Encapsulates the responsibility to load icons."""
@@ -39,6 +57,8 @@ class Icons(object):
         'detach': ['udiskie-detach'],
         'quit': ['application-exit'],
         'forget_password': ['edit-delete'],
+        'delete': ['udiskie-eject'],
+        'losetup': ['udiskie-mount'],
     }
 
     def __init__(self, icon_names={}):
@@ -50,6 +70,20 @@ class Icons(object):
             if isinstance(v, basestring):
                 self._icon_names[k] = [v]
 
+    def get_icon_name(self, icon_id):
+        """
+        Try to determine the name of the icon that should be used.
+
+        :param str icon_id: udiskie internal icon id
+        :returns: the icon name to be used
+        :rtype: str
+        """
+        icon_theme = Gtk.IconTheme.get_default()
+        for name in self._icon_names[icon_id]:
+            if icon_theme.has_icon(name):
+                return name
+        return 'not-available'
+
     def get_icon(self, icon_id, size):
         """
         Load icon dynamically.
@@ -82,6 +116,7 @@ class UdiskieMenu(object):
     """
 
     _quit_label = _('Quit')
+    _losetup_label = _('Setup loop device')
 
     def __init__(self, mounter, icons, actions, quit_action=None):
         """
@@ -113,15 +148,22 @@ class UdiskieMenu(object):
         self._actions = actions
         self._quit_action = quit_action
 
-    def __call__(self):
+    def __call__(self, menu):
         """
-        Create menu for udiskie mount operations.
+        Populate the menu with udiskie mount operations.
 
-        :returns: a new menu
-        :rtype: Gtk.Menu
+        :param Gtk.Menu menu:
         """
         # create actions items
-        menu = self._branchmenu(self._prepare_menu(self.detect()).groups)
+        self._create_menu_items(menu, self._prepare_menu(self.detect()))
+        if self._mounter.udisks.loop_support:
+            if len(menu) > 0:
+                menu.append(Gtk.SeparatorMenuItem())
+            menu.append(self._menuitem(
+                self._losetup_label,
+                self._icons.get_icon('losetup', Gtk.IconSize.MENU),
+                lambda _: self._losetup()
+            ))
         # append menu item for closing the application
         if self._quit_action:
             if len(menu) > 0:
@@ -133,6 +175,20 @@ class UdiskieMenu(object):
             ))
         return menu
 
+    @Coroutine.from_generator_function
+    def _losetup(self):
+        dialog = Gtk.FileChooserDialog(
+            _('Open disc image'), None,
+            Gtk.FileChooserAction.OPEN,
+            (_('Open'), Gtk.ResponseType.OK,
+             _('Cancel'), Gtk.ResponseType.CANCEL))
+        dialog.show_all()
+        response = yield Dialog(dialog)
+        dialog.hide()
+        if response != Gtk.ResponseType.OK:
+            return
+        yield self._mounter.losetup(dialog.get_filename())
+
     def detect(self):
         """
         Detect all currently known devices.
@@ -140,39 +196,49 @@ class UdiskieMenu(object):
         :returns: root of device hierarchy
         :rtype: Device
         """
-        return self._actions.detect()
+        root = self._actions.detect()
+        prune_empty_node(root, set())
+        return root
 
-    def _branchmenu(self, groups):
+    def _create_menu(self, items):
         """
         Create a menu from the given node.
 
-        :param Branch groups: contains information about the menu
-        :returns: a new menu object holding all groups of the node
+        :param list items: list of menu items
+        :returns: a new menu object holding all items of the node
         :rtype: Gtk.Menu
         """
+        menu = Gtk.Menu()
+        self._create_menu_items(menu, items)
+        return menu
+
+    def _create_menu_items(self, menu, items):
         def make_action_callback(node):
             return lambda _: node.action()
-        menu = Gtk.Menu()
-        separate = False
-        for group in groups:
-            if len(group) > 0:
-                if separate:
-                    menu.append(Gtk.SeparatorMenuItem())
-                separate = True
-            for node in group:
-                if isinstance(node, Action):
-                    menu.append(self._menuitem(
-                        node.label,
-                        self._icons.get_icon(node.method, Gtk.IconSize.MENU),
-                        make_action_callback(node)))
-                elif isinstance(node, Branch):
-                    menu.append(self._menuitem(
-                        node.label,
-                        icon=None,
-                        onclick=self._branchmenu(node.groups)))
-                else:
-                    raise ValueError(_("Invalid node!"))
-        return menu
+        for node in items:
+            if isinstance(node, Action):
+                menu.append(self._menuitem(
+                    node.label,
+                    self._icons.get_icon(node.method, Gtk.IconSize.MENU),
+                    make_action_callback(node)))
+            elif isinstance(node, SubMenu):
+                menu.append(self._menuitem(
+                    node.label,
+                    icon=None,
+                    onclick=self._create_menu(node.items)))
+            elif isinstance(node, MenuSection):
+                self._create_menu_section(menu, node)
+            else:
+                raise ValueError(_("Invalid node!"))
+
+    def _create_menu_section(self, menu, section):
+        if len(menu) > 0:
+            menu.append(Gtk.SeparatorMenuItem())
+        if section.label:
+            mi = self._menuitem(section.label, None, None)
+            mi.set_sensitive(False)
+            menu.append(mi)
+        self._create_menu_items(menu, section.items)
 
     def _menuitem(self, label, icon, onclick):
         """
@@ -196,7 +262,7 @@ class UdiskieMenu(object):
             item.set_label(label)
         if isinstance(onclick, Gtk.Menu):
             item.set_submenu(onclick)
-        else:
+        elif onclick is not None:
             item.connect('activate', onclick)
         return item
 
@@ -206,63 +272,63 @@ class UdiskieMenu(object):
 
         :param Device node: root node of device hierarchy
         :returns: menu hierarchy
-        :rtype: Branch
+        :rtype: list
         """
-        return Branch(
-            label=node.label,
-            groups=[
-                [self._prepare_menu(branch)
-                 for branch in node.branches
-                 if branch.methods or branch.branches],
-                node.methods,
-            ])
+        return [
+            MenuSection(None, [
+                SubMenu(branch.label, self._prepare_menu(branch))
+                for branch in node.branches
+                if branch.methods or branch.branches
+            ]),
+            MenuSection(None, node.methods),
+        ]
 
 
 class SmartUdiskieMenu(UdiskieMenu):
 
-    def _actions_group(self, node, presentation):
-        """
-        Create the actions group for the specified device node.
-
-        :param Device node: device
-        :param str presentation: node label
-        """
-        labels = self._actions._labels
-        return [Action(action.method,
-                       action.device,
-                       labels[action.method].format(presentation),
-                       action.action)
-                for action in node.methods]
-
-    def _collapse_device(self, node, presentation=""):
+    def _collapse_device(self, node):
         """Collapse device hierarchy into a flat folder."""
-        if (not presentation
-                or node.device.is_mounted
-                or not node.device.is_luks_cleartext):
-            presentation = node.label
-        groups = [group
-                  for branch in node.branches
-                  for group in self._collapse_device(branch, presentation)
-                  if group]
-        groups.append(self._actions_group(node, presentation))
-        return groups
+        items = [item
+                 for branch in node.branches
+                 for item in self._collapse_device(branch)
+                 if item]
+        items.append(MenuSection(None, node.methods))
+        return items
 
     def _prepare_menu(self, node):
         """Overrides UdiskieMenu._prepare_menu."""
-        return Branch(
-            label=node.label,
-            groups=[
-                [Branch(branch.label, self._collapse_device(branch))
+        return [
+            SubMenu(branch.label, self._collapse_device(branch))
+            for branch in node.branches
+            if branch.methods or branch.branches
+        ]
+
+
+class FlatUdiskieMenu(UdiskieMenu):
+
+    def _collapse_device(self, node):
+        """Collapse device hierarchy into a flat folder."""
+        items = [item
                  for branch in node.branches
-                 if branch.methods or branch.branches],
-            ])
+                 for item in self._collapse_device(branch)
+                 if item]
+        items.extend(node.methods)
+        return items
+
+    def _prepare_menu(self, node):
+        """Overrides UdiskieMenu._prepare_menu."""
+        return [
+            MenuSection(branch.label, self._collapse_device(branch))
+            for branch in node.branches
+            if branch.methods or branch.branches
+        ]
 
 
 class TrayIcon(object):
 
     """Default TrayIcon class."""
 
-    def __init__(self, menumaker, icons, statusicon=None, show=True):
+    def __init__(self, menumaker, icons, statusicon=None):
         """
         Create an object managing a tray icon.
 
@@ -281,8 +347,6 @@ class TrayIcon(object):
         self._conn_right = None
         self.task = Async()
         menumaker._quit_action = self.destroy
-        if show:
-            self.show()
 
     def destroy(self):
         self.show(False)
@@ -326,7 +390,9 @@ class TrayIcon(object):
 
     def create_context_menu(self):
         """Create the context menu."""
-        return self._menu()
+        menu = Gtk.Menu()
+        self._menu(menu)
+        return menu
 
     def _activate(self, icon):
         """Handle a left click event (show the menu)."""
@@ -346,34 +412,47 @@ class TrayIcon(object):
         self._m = m
 
 
-class AutoTray(TrayIcon):
+class UdiskieStatusIcon(object):
 
     """
-    TrayIcon that automatically hides.
+    Manage a status icon.
 
-    The menu has no 'Quit' item, and the tray icon will automatically hide
-    if there is no action available.
+    When `smart` is on, the icon will automatically hide if there is no action
+    available and the menu will have no 'Quit' item.
     """
 
-    def __init__(self, menumaker, icons):
-        """
-        Create and automatically set visibility of a new status icon.
-
-        Overrides TrayIcon.__init__.
-        """
-        super(AutoTray, self).__init__(menumaker, icons, show=False)
-        # Okay, the following is BAD:
-        menumaker._quit_action = None
+    def __init__(self, icon, menumaker, smart=False):
+        self._icon = icon
+        self._menumaker = menumaker
+        self._quit_action = menumaker._quit_action
         udisks = menumaker._mounter.udisks
         udisks.connect('device_changed', self.update)
         udisks.connect('device_added', self.update)
         udisks.connect('device_removed', self.update)
+        self.smart = smart
+
+    @property
+    def smart(self):
+        return getattr(self, '_smart', None)
+
+    @smart.setter
+    def smart(self, smart):
+        if smart == self.smart:
+            return
+        if smart:
+            self._menumaker._quit_action = None
+        else:
+            self._menumaker._quit_action = self._quit_action
+        self._smart = smart
         self.update()
 
     def has_menu(self):
         """Check if a menu action is available."""
-        return any(self._menu._prepare_menu(self._menu.detect()).groups)
+        return any(self._menumaker._prepare_menu(self._menumaker.detect()))
 
     def update(self, *args):
         """Show/hide icon depending on whether there are devices."""
-        self.show(self.has_menu())
+        if self.smart:
+            self._icon.show(self.has_menu())
+        else:
+            self._icon.show(True)
